"""withdrawal Module."""


class withdrawal(object):
    """withdrawal methods."""

    def __init__(self, main):
        """Initialise with super's main."""
        self.main = main

    def list_withdrawal_requests(self):
        """Get list of withdrawal requests."""
        trades = self.main.api_request('withdrawals')
        return trades

    def create_withdrawal_request(self, wtype, amount, beneficiary_id=None):
        """Create a new withdrawal request."""
        data = {
            'type': wtype,
            'amount': amount,
            'beneficiary_id': beneficiary_id,
        }
        result = self.main.api_request('withdrawals',
                                       data=data, http_call='post')
        return result

    def get_withdrawals_status(self, wid=None):
        """Get status od specific withdrawal.

        :param wid: String.
        :return:
        """
        data = {'id': wid}
        return self.main.api_request('withdrawals', params=data)

    def delete_withdrawal_request(self, wid):
        """Delete a new withdrawal request."""
        data = {'id': wid}
        result = self.main.api_request('withdrawals',
                                       params=data, http_call='delete')
        return result





import functools
import inspect
import logging
import traceback
import warnings
# class BaseClass(object):
#     def __init__(self):
from time import sleep, time

log = logging.getLogger(__name__)



def RateLimiter(f):
    """Rate Limiter decorator."""
    burst_count, first_burst = ([0], [0.0])

    def wrapper(*args, **kwargs):
        max_rate = args[0].maxRate
        burst = args[0].maxBurst
        cur_time = time()
        if (max_rate is None) or (burst is None):
            return f(*args, **kwargs)
            # return ret
        min_interval = 1.0 / float(max_rate)
        if burst_count[0] == 0:
            first_burst[0] = cur_time
            burst_count[0] += 1
        if burst_count[0] <= burst:
            burst_count[0] += 1
            # ret = f(*args, **kwargs)
        if burst_count[0] == burst+1:
            elapsed = cur_time - first_burst[0]
            # ret = f(*args, **kwargs)
            burst_count[0] = 1
            wait_time = burst*min_interval - elapsed
            if wait_time > 0:
                log.warning('Rate limited! Waiting {:.2f}s'.format(wait_time))
                sleep(wait_time)
        if burst_count[0] == 1:
            first_burst[0] = time()
        return f(*args, **kwargs)
    return wrapper


class LunoAPIError(ValueError):
    """Generic Error Class."""

    def __init__(self, response):
        """Instatiate with an error response."""
        self.url = response.url
        self.code = response.status_code
        self.message = response.text

    def __str__(self):
        """Return a string error message."""
        return "Luno request %s failed with %d: %s" % (
            self.url, self.code, self.message)


class LunoAPIRateLimitError(ValueError):
    """Generic rate limit error class."""

    def __init__(self, response):
        """Instanciate with a rate error response."""
        self.url = response.url
        self.code = response.status_code
        self.message = response.text

    def __str__(self):
        """Return a string error message."""
        return "Rate Limit Error.\nLuno request %s failed with %d: %s" % (
            self.url, self.code, self.message)



"""Markets module."""
import logging

import pandas as pd

log = logging.getLogger(__name__)


class Market(object):
    """Market related methods."""

    def __init__(self, main):
        """Initialise with super's main."""
        self.main = main

    def get_ticker(self, kind='auth', pair=None):
        """Get the latest ticker indicator for a pair."""
        params = {'pair': self.main.pair if pair is None else pair}
        return self.main.api_request('ticker', params, kind=kind)

    def get_all_tickers(self, kind='auth'):
        """Get all the latest ticker indicators."""
        return self.main.api_request('tickers', None, kind=kind)

    def get_order_book(self, limit=None, kind='auth', pair=None):
        """Get a list of bids and asks in the order book."""
        params = {'pair': self.main.pair if pair is None else pair}
        orders = self.main.api_request('orderbook', params, kind=kind)
        if limit is not None:
            orders['bids'] = orders['bids'][:limit]
            orders['asks'] = orders['asks'][:limit]
        return orders

    def get_order_book_frame(self, limit=None, kind='auth', pair=None):
        """Get orderbook as a dataframe."""
        q = self.get_order_book(limit, kind, pair)
        asks = pd.DataFrame(q['asks'])
        bids = pd.DataFrame(q['bids'])
        index = pd.MultiIndex.from_product(
            [('asks', 'bids'), ('price', 'volume')])
        df = pd.DataFrame(
            pd.concat([asks, bids], axis=1).values, columns=index)
        return df

    def get_trades(self, limit=None, kind='auth', since=None, pair=None):
        """Get a list of the most recent trades."""
        params = {'pair': self.main.pair if pair is None else pair}
        if since is not None:
            params['since'] = since
        trades = self.main.api_request('trades', params, kind=kind)
        if limit is not None:
            trades['trades'] = trades['trades'][:limit]
        return trades

    def get_trades_frame(self, limit=None, kind='auth', since=None, pair=None):
        """Get a dataframe of the most recent trades."""
        trades = self.get_trades(limit, kind, since, pair)
        df = pd.DataFrame(trades['trades'])
        if not df.empty:
            df.index = pd.to_datetime(df.timestamp, unit='ms')
            df.price = df.price.apply(pd.to_numeric)
            df.volume = df.volume.apply(pd.to_numeric)
            df.drop('timestamp', axis=1, inplace=True)
        else:
            log.warning('Empty response from get_trades. Returning empty df')
        return df



__version__ = "0.3.0"

"""
Changelog:
0.3.0- Changed call structure- separated things a bit.
0.2.1- Added some of the missing tests. Updated readmes.
0.2.0- Name change and added endpoints market orders, make account,
    plus various other bug fixes. Also added rate limiting fuctionality.

"""



"""Base API Module."""
from __future__ import absolute_import

import logging
from concurrent.futures import ThreadPoolExecutor

import requests

from . import meta
from .accounts import Account
from .market import Market
from .orders import Orders
from .quotes import Quotes
from .receive import Receive
from .utils import LunoAPIError, LunoAPIRateLimitError, RateLimiter
from .withdrawal import withdrawal

__version__ = meta.__version__

log = logging.getLogger(__name__)


class Luno(object):
    """Main Luno API class."""

    def __init__(self, key, secret, options={}):
        """Instantiate with key and secret if authentication is wanted."""
        self.options = options
        self.auth = (key, secret)
        if 'hostname' in options:
            self.hostname = options['hostname']
        else:
            self.hostname = 'api.mybitx.com'
        self.port = options['port'] if 'port' in options else 443
        self.pair = options['pair'] if 'pair' in options else 'XBTZAR'
        self.ca = options['ca'] if 'ca' in options else None
        self.timeout = options['timeout'] if 'timeout' in options else 30
        self.maxRate = options['maxRate'] if 'maxRate' in options else 0.1
        self.maxBurst = options['maxBurst'] if 'maxBurst' in options else 5
        self.headers = {
            'Accept': 'application/json',
            'Accept-Charset': 'utf-8',
            'User-Agent': 'py-luno v' + __version__
        }
        # Use a Requests session so that we can keep headers and connections
        # across API requests
        self._requests_session = requests.Session()
        self._requests_session.headers.update(self.headers)
        self._executor = ThreadPoolExecutor(max_workers=5)

        self.account = Account(self)
        self.market = Market(self)
        self.orders = Orders(self)
        self.quotes = Quotes(self)
        self.receive = Receive(self)
        self.withdrawal = withdrawal(self)

    def close(self):
        """Close connection."""
        log.info('Asking MultiThreadPool to shutdown')
        self._executor.shutdown(wait=True)
        log.info('MultiThreadPool has shutdown')

    def construct_url(self, call):
        """Construc API Url."""
        base = self.hostname
        if self.port != 443:
            base += ':%d' % (self.port,)
        return "https://%s/api/1/%s" % (base, call)

    @RateLimiter
    def api_request(self, call, params=None, data=None,
                    kind='auth', http_call='get'):
        """General API request.

        Generally, use the convenience functions below
        :param kind: the type of request to make. 'auth' makes an
            authenticated call; 'basic' is unauthenticated
        :param call: the API call to make
        :param params: a dict of query parameters
        :return: a json response, a LunoAPIError is thrown if
            the api returns with an error
        """
        url = self.construct_url(call)
        auth = self.auth if kind == 'auth' else None
        response = self._requests_session.request(
            http_call.upper(),
            url, params=params, data=data, auth=auth, timeout=self.timeout)
        try:
            result = response.json()
        except ValueError:
            result = {'error': 'No JSON content returned'}
        if response.status_code in [429, 503]:
            log.error('Rate Limit Exceeded')
            raise LunoAPIRateLimitError(response)
        elif response.status_code != 200 or 'error' in result:
            raise LunoAPIError(response)
        else:
            return result

    def send_bitcoin(self, amount, currency, address,
                     description=None, message=None):
        """Send currency to account."""
        data = {
            'amount': amount,
            'currency': currency,
            'address': address,
            'description': description,
            'message': message,
        }
        result = self.api_request('send', params=data, http_call='post')
        return result

    # @deprecated('Use account.create_account instead')
    # def create_account(self, *args, **kwargs):
    #     return self.account.create_account(args, kwargs)
    #
    # @deprecated('Use account.get_balance instead')
    # def get_balance(self):
    #     return self.account.get_balance()
    #
    # @deprecated('Use account.get_transactions instead')
    # def get_transactions(self, *args, **kwargs):
    #     return self.account.get_transactions(args, kwargs)
    #
    # @deprecated('Use account.get_transactions_frame instead')
    # def get_transactions_frame(self, *args, **kwargs):
    #     return self.account.get_transactions_frame(args, kwargs)
    #
    # @deprecated('Use account.get_pending_transactions instead')
    # def get_pending_transactions(self, *args, **kwargs):
    #     return self.account.get_pending_transactions(args, kwargs)
    #
    # @deprecated('Use account.get_orders instead')
    # def get_orders(self, *args, **kwargs):
    #     return self.account.get_orders(args, kwargs)
    #
    # @deprecated('Use account.get_orders instead')
    # def get_orders(self, *args, **kwargs):
    #     return self.account.get_orders(args, kwargs)



"""Receive Module."""


class Receive(object):
    """Receive related methods."""

    def __init__(self, main):
        """Initialise with super's main."""
        self.main = main

    def get_receive_address(self, asset, address=None, kind='auth'):
        """Get receive address for the account.

        Returns the default receive address associated with your account
        and the amount received via the address. You can specify an
        optional address parameter to return information for a
        non-default receive address. In the response, total_received
        is the total confirmed Bitcoin amount received excluding
        unconfirmed transactions. total_unconfirmed is the total
        sum of unconfirmed receive transactions.
        """
        params = {'asset': asset,
                  'address': address}
        return self.main.api_request('funding_address', params, kind=kind)

    def create_receive_address(self, asset='XBT'):
        """Create a receive address.

        Allocates a new receive address to your account.
        There is a rate limit of 1 address per hour,
        but bursts of up to 10 addresses are allowed.
        """
        data = {
            'asset': asset,
        }
        result = self.main.api_request('funding_address',
                                       data=data, http_call='post')
        return result



"""Accounts Module."""
import json

import pandas as pd


class Account(object):
    """Class with the account methods."""

    def __init__(self, main):
        """Initialise with super's main."""
        self.main = main

    def create_account(self, currency, name, base_account_id, counter_id):
        """Create a new account in the selected currency.

        :param currency: Currency of account
        :param name: Name of account
        :param base_account_id: Id of the base account
        :param counter_id: Id of the counter account
        :return: dict with name. currency and id of new account
        """
        data = {
            'currency': currency,
            'name': name,
            'base_account_id': base_account_id,
            'counter_id': counter_id
        }
        return self.main.api_request('accounts', data=data, http_call='post')

    def get_balance(self):
        """Get balances of all accounts."""
        return self.main.api_request('balance', None)

    def get_transactions(self, account_id, min_row=None, max_row=None):
        """Get list of transactions for an account."""
        params = {}
        if min_row is not None:
            params['min_row'] = min_row
        if max_row is not None:
            params['max_row'] = max_row
        return self.main.api_request(
            'accounts/%s/transactions' % (account_id,), params)

    def get_transactions_frame(self, account_id, min_row=None, max_row=None):
        """Get dataframe of transactions for an account."""
        tx = self.get_transactions(
            account_id, min_row, max_row)['transactions']
        df = pd.DataFrame(tx)
        df.index = pd.to_datetime(df.timestamp, unit='ms')
        df.drop('timestamp', axis=1, inplace=True)
        return df

    def get_pending_transactions(self, account_id):
        """Get a list of pending transactions for an account."""
        return self.main.api_request(
            'accounts/%s/pending' % (account_id,), None)

    def get_orders(self, state=None, pair=None):
        """Get a list of most recently placed orders.

        You can specify an optional state='PENDING' parameter to
        restrict the results to only open orders. You can also specify the
            market by using the optional pair parameter.
        The list is truncated after 100 items.
        :param kind: typically 'auth' if you want this to return anything
            useful
        :param state: String optional 'COMPLETE', 'PENDING', or None (default)
        :return:
        """
        params = {'pair': self.main.pair if pair is None else pair}
        if state is not None:
            params['state'] = state
        return self.main.api_request('listorders', params)

    def get_orders_frame(self, state=None, kind='auth', pair=None):
        """Get a list of most recently placed orders as a dataframe."""
        q = self.get_orders(state, pair)
        tj = json.dumps(q['orders'])
        df = pd.read_json(
            tj, convert_dates=['creation_timestamp', 'expiration_timestamp'])
        df.index = df.creation_timestamp
        return df

    def create_transfer(self, amount, currency, note,
                        source_account_id, target_account_id):
        """Transfer currency between accounts."""
        data = {
            'amount': amount,
            'currency': currency,
            'note': note,
            'source_account_id': source_account_id,
            'target_account_id': target_account_id,
        }
        return self.main.api_request('transfers', data=data,
                                     http_call='post')

    def get_transfers(self, tid=None):
        """Get list of transfers."""
        return self.main.api_request('transfers/{}'.format(tid or ''),
                                     http_call='get')

    def confirm_transfer(self, tid):
        """Confirm a pending transfer."""
        return self.main.api_request('transfers/{}'.format(tid),
                                     http_call='put')



from __future__ import print_function

import os

from pyluno.api import Luno


def runDemo():
    user = ''
    password = ''
    auth = 'BITX_KEY' in os.environ and 'BITX_SECRET' in os.environ
    if auth:
        user = os.environ['BITX_KEY']
        password = os.environ['BITX_SECRET']
    else:
        print("Note: Couldn't find a BITX_KEY environment variable."
              " This means that none of the API queries that"
              " require authentication will work."
              " Carrying on anyway, but make sure your"
              " credentials are available "
              " in the BITX_KEY and BITX_SECRET environment variables"
              " and run this demo again")
    api = Luno(user, password)
    print(api.market.get_ticker())
    print(api.market.get_all_tickers())
    print(api.market.get_order_book(5))
    print(api.market.get_trades(10))
    if auth:
        print(api.account.get_orders())
        print(api.withdrawal.get_funding_address('XBT'))
        print(api.account.get_balance())


if __name__ == '__main__':
    runDemo()



"""Quotes module."""


class Quotes(object):
    """Class for quote related methods."""

    def __init__(self, main):
        """Initialise with super's main."""
        self.main = main

    def get_quote(self, ttype, base_amount, pair):
        """Get temporary quote."""
        data = {
            'type': ttype,
            'base_amount': base_amount,
            'pair': pair,
        }
        result = self.main.api_request('quotes', data=data, http_call='post')
        return result

    def get_quote_status(self, wid):
        """Get status of a specific quote."""
        data = {'id': wid}
        result = self.main.api_request('quotes',
                                       params=data)
        return result

    def execute_quote(self, wid):
        """Execute trade based on quote."""
        data = {'id': wid}
        result = self.main.api_request('quotes', params=data, http_call='put')
        return result

    def delete_quote(self, wid):
        """Delete quote."""
        data = {'id': wid}
        result = self.main.api_request('quotes', params=data,
                                       http_call='delete')
        return result



"""Orders Module."""
import logging

import pandas as pd

log = logging.getLogger(__name__)


class Orders(object):
    """Class with order related methods."""

    def __init__(self, main):
        """Initialise with super's main."""
        self.main = main

    def create_limit_order(self, order_type, volume, price,
                           base_account_id, counter_account_id):
        """Create a new limit order.

        :param order_type: 'buy' or 'sell'
        :param volume: the volume, in BTC
        :param price: the ZAR price per bitcoin
        :return: the order id
        """
        data = {
            'pair': self.main.pair,
            'type': 'BID' if order_type == 'buy' else 'ASK',
            'volume': str(volume),
            'price': str(price),
            'base_account_id': base_account_id,
            'counter_account_id': counter_account_id,
        }
        result = self.main.api_request('postorder', data=data,
                                       http_call='post')
        return result

    def create_market_order(self, order_type, volume,
                            base_account_id, counter_account_id):
        """Create a new market order.

        :param order_type: 'buy' or 'sell'
        :param volume: the volume of btc if sell, or currency if buy.
        :return: the order id
        """
        data = {
            'pair': self.main.pair,
            'type': 'BUY' if order_type == 'buy' else 'SELL',
            'volume': str(volume),
            'base_account_id': base_account_id,
            'counter_account_id': counter_account_id,
        }
        if order_type is 'buy':
            data['couter_volume'] = volume
        else:
            data['base_volume'] = volume
        result = self.main.api_request('marketorder', data=data,
                                       http_call='post')
        return result

    def stop_order(self, order_id):
        """Stop a specific order.

        :param order_id: The order ID
        :return: a success flag
        """
        data = {
            'order_id': order_id,
        }
        return self.main.api_request('stoporder', data=data,
                                     http_call='post')

    def stop_all_orders(self):
        """Stop all pending orders, both sell and buy.

        :return: dict of Boolean -- whether request succeeded or not for each
            order_id that was pending
        """
        pending = self.main.get_orders('PENDING')['orders']
        ids = [order['order_id'] for order in pending]
        result = {}
        for order_id in ids:
            status = self.main.stop_order(order_id)
            result[order_id] = status['success']
        return result

    def get_order(self, order_id):
        """Get an order by its ID.

        :param order_id: string	The order ID
        :return: dict order details or LunoAPIError raised
        """
        return self.main.api_request('orders/{}'.format(order_id))

    def list_trades(self, limit=None, since=None, pair=None):
        """Get list of all trades."""
        params = {
            'since': since,
            'limit': limit,
        }
        params = {'pair': self.main.pair if pair is None else pair}
        trades = self.main.api_request('listtrades', params)
        return trades

    def list_trades_frame(self, limit=None, since=None, pair=None):
        """Get dataframe of all trades."""
        trades = self.list_trades(limit, since, pair)
        df = pd.DataFrame(trades['trades'])
        if not df.empty:
            df.index = pd.to_datetime(df.timestamp, unit='ms')
            df.price = df.price.apply(pd.to_numeric)
            df.volume = df.volume.apply(pd.to_numeric)
            df.base = df.base.apply(pd.to_numeric)
            df.counter = df.counter.apply(pd.to_numeric)
            df.fee_base = df.fee_base.apply(pd.to_numeric)
            df.drop('timestamp', axis=1, inplace=True)
        else:
            log.warning('Empty response from list_trades. Returning empty df')
        return df

    def get_fee_info(self, kind='auth', pair=None):
        """Get the fee info for the account."""
        params = {'pair': self.main.pair if pair is None else pair}
        return self.main.api_request('fee_info', params, kind=kind)



from setuptools import setup, find_packages
with open('pyluno/meta.py') as f:
    exec(f.read())

setup(
    name='pyluno',
    version=__version__,
    packages=find_packages(exclude=['tests']),
    description='A Luno API for Python',
    author='Cayle Sharrock/Grant Stephens',
    author_email='grant@stephens.co.za',
    scripts=['demo.py'],
    install_requires=[
        'futures>=3.0.3',
        'nose>=1.3.7',
        'requests>=2.8.1',
        'pandas>=0.17.0',
    ],
    license='MIT',
    url='https://github.com/grantstephens/pyluno',
    download_url='https://github.com/grantstephens/pyluno/tarball/%s'
        % (__version__, ),
    keywords='Luno Bitcoin exchange API',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Topic :: Office/Business :: Financial',
        'Topic :: Utilities',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    test_suite='tests',
    extras_require={
        'test':  ['requests-mock>=0.7.0', 'nose'],
        }
)
